///////////////////////////////////////////////////////////////////////////////
//
//   PINOLVXD.C -- VXD For PinoLite On Win95
//
///////////////////////////////////////////////////////////////////////////////
#include "pch.h"
#include <vwin32.h>
#include <winerror.h>
#include <vmm.h>

#include "..\include\pcicnfg.h"
#include "..\include\PinoLCtl.h"

#pragma warning (disable:4100)  // Parameter unused

#pragma CM_PAGEABLE_DATA

// Control Data
typedef struct CTL_DATA_t {
    DEVNODE devnode;            // Device Node
    PCICONFIG PciConfig;        // PCI Configration
    DWORD Pci_IntAdrs;          // PCI BUS OPERATION REGISTERS Physical Int Base Address
    DWORD Pci_ExtAdrs;          // PCI BUS OPERATION REGISTERS Physical Ext Base Address
    DWORD Pci_IO1Adrs;          // PCI BUS OPERATION REGISTERS IO1 Base Address
    DWORD Pci_IO2Adrs;          // PCI BUS OPERATION REGISTERS IO2 Base Address
    DWORD Pci_IntLinAdrs;       // PCI BUS OPERATION REGISTERS Int Linear Base Address
    DWORD Pci_ExtLinAdrs;       // PCI BUS OPERATION REGISTERS Ext Linear Base Address
} CTL_DATA;

CTL_DATA CtlData;

#pragma CM_PAGEABLE_CODE


CONFIGRET OnConfigStart( SUBCONFIGFUNC scf, DEVNODE devnode, DWORD refdata, ULONG flags);
CONFIGRET OnConfigStop( SUBCONFIGFUNC scf, DEVNODE devnode, DWORD refdata, ULONG flags);

static CONFIGRET GetRegistryValue( DEVNODE dnDevNode, PFARCHAR pszValueName, PFARVOID Buffer, ULONG pulLength );
static void WriteConfigDWord( DEVNODE devnode, ULONG offset, DWORD data );

/*--- DeviceIoControl Function  ---*/
static int GetVersion( DIOCPARAMETERS *lpDIOCParms );
static int GetVendorID( DIOCPARAMETERS *lpDIOCParms ) ;
static int GetDeviceID( DIOCPARAMETERS *lpDIOCParms ) ;
static int GetPinoIntAddress( DIOCPARAMETERS *lpDIOCParms ) ;
static int GetPinoExtAddress( DIOCPARAMETERS *lpDIOCParms ) ;
static int GetPinoIntLinearAddress( DIOCPARAMETERS *lpDIOCParms ) ;
static int GetPinoExtLinearAddress( DIOCPARAMETERS *lpDIOCParms ) ;
static int GetPinoIO1Address( DIOCPARAMETERS *lpDIOCParms ) ;
static int GetPinoIO2Address( DIOCPARAMETERS *lpDIOCParms ) ;
static int LinToPhy( DIOCPARAMETERS *lpDIOCParms ) ;
static int PhyToLin( DIOCPARAMETERS *lpDIOCParms ) ;
static int AllocLinearPage( DIOCPARAMETERS *lpDIOCParms ) ;
static int FreeLinearPage( DIOCPARAMETERS *lpDIOCParms ) ;
static int GetPinoPCIConfig( DIOCPARAMETERS *lpDIOCParms ) ;
static int SetPinoBaseAddress( DIOCPARAMETERS *lpDIOCParms ) ;

// Ʊڽ -> 486CPU߰ðٴ(PTE)o
BOOL VXDINLINE VMM_CopyPageTable( ULONG LinPgNum, ULONG nPages, ULONG *PageBuf, ULONG flags )
{
    BOOL rc;
    Touch_Register(eax)
    Touch_Register(ecx)
    Touch_Register(edx)
    _asm push [flags]
    _asm push [PageBuf]
    _asm push [nPages]
    _asm push [LinPgNum]
    VMMCall(_CopyPageTable);
    _asm add  esp, 4*4
    _asm mov [rc], eax
    return(rc);
}

/****************************************************************************
 *  OnConfigure
 *
 ***************************************************************************/
CONFIGRET OnConfigure( CONFIGFUNC cf, SUBCONFIGFUNC scf,
						DEVNODE devnode, DWORD refdata, ULONG flags)
{
	switch ( cf ){
	case CONFIG_START:	// gpJn
		return ( OnConfigStart( scf, devnode, refdata, flags ));
	case CONFIG_STOP:	// gpI
	case CONFIG_REMOVE:
		return ( OnConfigStop( scf, devnode, refdata, flags ));
	default:
		return CR_DEFAULT;
	}
}

/****************************************************************************
 *  OnConfigStart
 *
 *
 ***************************************************************************/
CONFIGRET OnConfigStart( SUBCONFIGFUNC scf, DEVNODE devnode, DWORD refdata, ULONG flags)
{
    CtlData.PciConfig.vendorid = 0 ;
    CtlData.PciConfig.deviceid = 0 ;

    // Get PCI Configration
    CM_Call_Enumerator_Function( devnode, PCI_ENUM_FUNC_GET_DEVICE_INFO,
        0, &CtlData.PciConfig, sizeof CtlData.PciConfig, 0 );

    // PCI BUS OPERATION REGISTERS Int Base Address
    CtlData.Pci_IntAdrs = (CtlData.PciConfig.baseaddr[ 2 ] & 0xfffffff0) ;
    // PCI BUS OPERATION REGISTERS Ext Base Address
    CtlData.Pci_ExtAdrs = (CtlData.PciConfig.baseaddr[ 0 ] & 0xfffffff0) ;
    // PCI BUS OPERATION REGISTERS IO1 Base Address
    CtlData.Pci_IO1Adrs = (CtlData.PciConfig.baseaddr[ 4 ] & 0xfffffffc) ;
    // PCI BUS OPERATION REGISTERS IO2 Base Address
    CtlData.Pci_IO2Adrs = (CtlData.PciConfig.baseaddr[ 5 ] & 0xfffffffc) ;

    // DEVNODE Keep
    CtlData.devnode = devnode;

    CtlData.Pci_IntLinAdrs = (DWORD)_MapPhysToLinear(CtlData.Pci_IntAdrs,((INTADDRESSEND - INTADDRESSTOP) + 1) , 0 );
    CtlData.Pci_ExtLinAdrs = (DWORD)_MapPhysToLinear(CtlData.Pci_ExtAdrs,((EXTADDRESSEND - EXTADDRESSTOP) + 1) , 0 );

    return CR_SUCCESS;
}

/****************************************************************************
 *  OnConfigStop
 *
 *
 ***************************************************************************/
CONFIGRET OnConfigStop( SUBCONFIGFUNC scf, DEVNODE devnode, DWORD refdata, ULONG flags)
{
	return CR_SUCCESS;
}

/****************************************************************************
 *  OnDeviceIOControl
 *
 *  DeviceIoControl Handler
 *
 ***************************************************************************/
int CM_SYSCTRL OnDeviceIOControl( DWORD dwService, DWORD  dwDDB, DWORD a, DIOCPARAMETERS *lpDIOCParms )
{
    switch ( dwService ) {
    case DIOC_GETVERSION:
    case DIOC_CLOSEHANDLE:
        return ( 0 );
    case _GETVERSION:           // GetVersion
        return ( GetVersion( lpDIOCParms ));
    case _GETVENDORID:          // GetVendorID
        return ( GetVendorID( lpDIOCParms ));
    case _GETDEVICEID:          // GetDeviceID
        return ( GetDeviceID( lpDIOCParms ));
    case _GETPINOINTADDRESS:    // Get Pino Interanl Address
        return ( GetPinoIntAddress( lpDIOCParms ));
    case _GETPINOEXTADDRESS:    // Get Pino External Address
        return ( GetPinoExtAddress( lpDIOCParms ));
    case _GETPINOINTLINADDRESS:    // Get Pino Interanl Linear Address
        return ( GetPinoIntLinearAddress( lpDIOCParms ));
    case _GETPINOEXTLINADDRESS:    // Get Pino External Linear Address
        return ( GetPinoExtLinearAddress( lpDIOCParms ));
    case _GETPINOIO1ADDRESS:    // Get Pino IO Space 1 Address
        return ( GetPinoIO1Address( lpDIOCParms ));
    case _GETPINOIO2ADDRESS:    // Get Pino IO Space 21 Address
        return ( GetPinoIO2Address( lpDIOCParms ));
    case _LINEARTOPHY:          // Linear -> Physical Address
        return ( LinToPhy( lpDIOCParms ));
    case _PHYTOLINEAR:          // Physical -> Linear Address
        return ( PhyToLin( lpDIOCParms ));
    case _ALLOCLINEAR:          // Alloc Linear Memory Page
        return ( AllocLinearPage( lpDIOCParms ));
    case _FREELINEAR:           // Free Linear Memory Page
        return ( FreeLinearPage( lpDIOCParms ));
    case _GETPINOPCICONFIG:
        return( GetPinoPCIConfig( lpDIOCParms )) ;
	case _SETPINOBASEADDRESS:
        return( SetPinoBaseAddress( lpDIOCParms )) ;
    }

    /* Nothing Function */
    return ( 50 );	// M.S.J. No.39
}

/****************************************************************************
 *  Version
 *
 ***************************************************************************/
static int GetVersion( DIOCPARAMETERS *lpDIOCParms )
{
    if ( lpDIOCParms->cbOutBuffer >= sizeof (DWORD)){
        *(DWORD *)lpDIOCParms->lpvOutBuffer = 0x00010001;
        *(DWORD *)lpDIOCParms->lpcbBytesReturned = sizeof (DWORD);
        return ( 0 );
    } else {
        return ( -1 );
    }
}

/****************************************************************************
 *  PCI Vender ID
 *
 ***************************************************************************/
static int GetVendorID( DIOCPARAMETERS *lpDIOCParms )
{
    if ( lpDIOCParms->cbOutBuffer >= sizeof (DWORD)){
        *(DWORD *)lpDIOCParms->lpvOutBuffer = CtlData.PciConfig.vendorid ;
        *(DWORD *)lpDIOCParms->lpcbBytesReturned = sizeof (DWORD);
        return ( 0 );
    } else {
        return ( -1 );
    }
}
/****************************************************************************
 *  PCI Device ID
 *
 ***************************************************************************/
static int GetDeviceID( DIOCPARAMETERS *lpDIOCParms )
{
    if ( lpDIOCParms->cbOutBuffer >= sizeof (DWORD)){
        *(DWORD *)lpDIOCParms->lpvOutBuffer = CtlData.PciConfig.deviceid ;
        *(DWORD *)lpDIOCParms->lpcbBytesReturned = sizeof (DWORD);
        return ( 0 );
    } else {
        return ( -1 );
    }
}
/****************************************************************************
 *  Get Pino Lite Internal Address (Physical)
 *  Get Data
 ***************************************************************************/
static int GetPinoIntAddress( DIOCPARAMETERS *lpDIOCParms )
{
    if ( lpDIOCParms->cbOutBuffer >= sizeof (DWORD)){
        *(DWORD *)lpDIOCParms->lpvOutBuffer = CtlData.Pci_IntAdrs ;
        *(DWORD *)lpDIOCParms->lpcbBytesReturned = sizeof (DWORD) ;
        return ( 0 );
    } else {
        return ( -1 );
    }
}
/****************************************************************************
 *  Get Pino Lite External Address (Physical)
 *  Get Data
 ***************************************************************************/
static int GetPinoExtAddress( DIOCPARAMETERS *lpDIOCParms )
{
    if ( lpDIOCParms->cbOutBuffer >= sizeof (DWORD)){
        *(DWORD *)lpDIOCParms->lpvOutBuffer = CtlData.Pci_ExtAdrs ;
        *(DWORD *)lpDIOCParms->lpcbBytesReturned = sizeof (DWORD) ;
        return ( 0 );
    } else {
        return ( -1 );
    }
}
/****************************************************************************
 *  Get Pino Lite Internal Address (Linear)
 *  Get Data
 ***************************************************************************/
static int GetPinoIntLinearAddress( DIOCPARAMETERS *lpDIOCParms )
{
    if ( lpDIOCParms->cbOutBuffer >= sizeof (DWORD)){
        *(DWORD *)lpDIOCParms->lpvOutBuffer = CtlData.Pci_IntLinAdrs ;
        *(DWORD *)lpDIOCParms->lpcbBytesReturned = sizeof (DWORD) ;
        return ( 0 );
    } else {
        return ( -1 );
    }
}
/****************************************************************************
 *  Get Pino Lite External Address (Linear)
 *  Get Data
 ***************************************************************************/
static int GetPinoExtLinearAddress( DIOCPARAMETERS *lpDIOCParms )
{
    if ( lpDIOCParms->cbOutBuffer >= sizeof (DWORD)){
        *(DWORD *)lpDIOCParms->lpvOutBuffer = CtlData.Pci_ExtLinAdrs ;
        *(DWORD *)lpDIOCParms->lpcbBytesReturned = sizeof (DWORD) ;
        return ( 0 );
    } else {
        return ( -1 );
    }
}
/****************************************************************************
 *  Get Pino Lite IO Space 1 Address (Physical)
 *  Get Data
 ***************************************************************************/
static int GetPinoIO1Address( DIOCPARAMETERS *lpDIOCParms )
{
    if ( lpDIOCParms->cbOutBuffer >= sizeof (DWORD)){
        *(DWORD *)lpDIOCParms->lpvOutBuffer = CtlData.Pci_IO1Adrs ;
        *(DWORD *)lpDIOCParms->lpcbBytesReturned = sizeof (DWORD) ;
        return ( 0 );
    } else {
        return ( -1 );
    }
}
/****************************************************************************
 *  Get Pino Lite IO Space 2 Address (Physical)
 *  Get Data
 ***************************************************************************/
static int GetPinoIO2Address( DIOCPARAMETERS *lpDIOCParms )
{
    if ( lpDIOCParms->cbOutBuffer >= sizeof (DWORD)){
        *(DWORD *)lpDIOCParms->lpvOutBuffer = CtlData.Pci_IO2Adrs ;
        *(DWORD *)lpDIOCParms->lpcbBytesReturned = sizeof (DWORD) ;
        return ( 0 );
    } else {
        return ( -1 );
    }
}
/****************************************************************************
 *  Get Pino Lite PCI Configration Address
 *  Get Data
 ***************************************************************************/
static int GetPinoPCIConfig( DIOCPARAMETERS *lpDIOCParms )
{
    if ( lpDIOCParms->cbOutBuffer >= sizeof (DWORD)){
        *(DWORD *)lpDIOCParms->lpvOutBuffer = (DWORD)(&CtlData.PciConfig) ;
        *(DWORD *)lpDIOCParms->lpcbBytesReturned = sizeof (DWORD) ;
        return ( 0 );
    } else {
        return ( -1 );
    }
}
/****************************************************************************
 *  Linear -> Physical
 *
 ***************************************************************************/
static int LinToPhy( DIOCPARAMETERS *lpDIOCParms )
{
    if ( lpDIOCParms->cbOutBuffer >= sizeof (ULONG)){
        // jAy[Wԍy[We[uGg(PTE)擾
        ULONG lin = *((ULONG *)lpDIOCParms->lpvInBuffer);
        ULONG offset = lin % 4096;
        ULONG pgNum = lin / 4096;
        ULONG pte;
        if ( VMM_CopyPageTable( pgNum, 1, &pte, 0 )){
            // PTE ̉12bits 0ɂ̂AhX
            pte = pte / 4096 * 4096;
            *(ULONG *)lpDIOCParms->lpvOutBuffer = pte + offset;
            *(DWORD *)lpDIOCParms->lpcbBytesReturned = sizeof (ULONG);
            return ( 0 );
        } else {
            *(ULONG *)lpDIOCParms->lpvOutBuffer = 0;
            *(DWORD *)lpDIOCParms->lpcbBytesReturned = sizeof (ULONG);
        }
    }
    return ( -1 );
}

/****************************************************************************
 *  Phisical -> Linear
 *
 ***************************************************************************/
static int PhyToLin( DIOCPARAMETERS *lpDIOCParms )
{
    if ( lpDIOCParms->cbOutBuffer >= sizeof (void *)){
        PHY_TO_LIN_CMD *cmd = (PHY_TO_LIN_CMD *)lpDIOCParms->lpvInBuffer;
        *(PVOID *)lpDIOCParms->lpvOutBuffer = _MapPhysToLinear((ULONG)cmd->phyAddr, cmd->size, 0 );
        *(DWORD *)lpDIOCParms->lpcbBytesReturned = sizeof (ULONG);
        if ( *(PVOID *)lpDIOCParms->lpvOutBuffer != (PVOID)(-1)){
            // ϊ
            return ( 0 );
        }
    }
    return ( -1 );
}
/****************************************************************************
 *   Linear Memory Page Alloc
 *
 ***************************************************************************/
static int AllocLinearPage( DIOCPARAMETERS *lpDIOCParms )
{
    ULONG size = *((ULONG *)lpDIOCParms->lpvInBuffer);
    ULONG pageNum = (size + (PAGESIZE - 1)) / PAGESIZE;

    // A擾(VMM)
    PAGE_ALLOC_OUT *outP = (PAGE_ALLOC_OUT *)lpDIOCParms->lpvOutBuffer;
    outP->linAddr = _PageAllocate( pageNum, PG_SYS, 0, 0 /* 4K Align */,
                        0 /*minPhys*/, 0xfff /*maxPhys*/, &(outP->phyAddr),
                        PAGEUSEALIGN | PAGECONTIG | PAGEFIXED );


    if ( outP->linAddr == 0 ){
        // mێs
        return ( -1 );
    } else {
        return ( 0 );
    }
}


/****************************************************************************
 *  Set PINO Base Address
 *
 ***************************************************************************/
static int SetPinoBaseAddress( DIOCPARAMETERS *lpDIOCParms )
{
    PINOBASECFG *cmd = (PINOBASECFG *)lpDIOCParms->lpvInBuffer ;

    // PCI BUS OPERATION REGISTERS Int Base Address
    CtlData.Pci_IntAdrs = (cmd->Int_MemoryBase & 0xfffffff0) ;
    // PCI BUS OPERATION REGISTERS Ext Base Address
    CtlData.Pci_ExtAdrs = (cmd->Ext_MemoryBase & 0xfffffff0) ;
    // PCI BUS OPERATION REGISTERS IO1 Base Address
    CtlData.Pci_IO1Adrs = (cmd->IO0_Base & 0xfffffffc) ;
    // PCI BUS OPERATION REGISTERS IO2 Base Address
    CtlData.Pci_IO2Adrs = (cmd->IO1_Base & 0xfffffffc) ;

    CtlData.Pci_IntLinAdrs = (DWORD)_MapPhysToLinear(CtlData.Pci_IntAdrs,((INTADDRESSEND - INTADDRESSTOP) + 1) , 0 );
    CtlData.Pci_ExtLinAdrs = (DWORD)_MapPhysToLinear(CtlData.Pci_ExtAdrs,((EXTADDRESSEND - EXTADDRESSTOP) + 1) , 0 );

    return ( 0 );
}

/****************************************************************************
 *   Linear Memory Page Free
 *
 ***************************************************************************/
static int FreeLinearPage( DIOCPARAMETERS *lpDIOCParms )
{
    // mۂJ(VMM)
    DWORD linAddr = *((DWORD *)lpDIOCParms->lpvInBuffer);
    _PageFree((void *)linAddr, 0 );
    return ( 0 );
}

/****************************************************************************
 *  Get Registry Data
 *
 ***************************************************************************/
static CONFIGRET GetRegistryValue( DEVNODE dnDevNode, PFARCHAR pszValueName, PFARVOID Buffer, ULONG Length )
{
    CONFIGRET rc;
    rc = CM_Read_Registry_Value( dnDevNode, NULL, pszValueName, REG_BINARY, Buffer, &Length, CM_REGISTRY_SOFTWARE );
    if ( rc != CR_SUCCESS ){
        // default
        rc = CM_Write_Registry_Value( dnDevNode, NULL, pszValueName, REG_BINARY, Buffer, Length, CM_REGISTRY_SOFTWARE );
    }
    return ( rc );
}

/****************************************************************************
 *  Write PCI Configration Data
 *
 ***************************************************************************/
static void WriteConfigDWord( DEVNODE devnode, ULONG offset, DWORD data )
{
    CM_Call_Enumerator_Function( devnode, PCI_ENUM_FUNC_SET_DEVICE_INFO,
    offset, &data, sizeof data, 0 );
}

